/*
 * Filename:     ext2sys.c
 * Project:      ext2 file system driver for MiNT
 * 
 * Note:         Please send suggestions, patches or bug reports to me
 *               or the MiNT mailing list (mint@fishpool.com).
 * 
 * Copying:      Copyright 1999 Frank Naumann (fnaumann@cs.uni-magdeburg.de)
 *               Copyright 1998, 1999 Axel Kaiser (DKaiser@AM-Gruppe.de)
 * 
 * Portions copyright 1992, 1993, 1994, 1995 Remy Card (card@masi.ibp.fr)
 * and 1991, 1992 Linus Torvalds (torvalds@klaava.helsinki.fi)
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

# include "ext2sys.h"
# include "ext2dev.h"
# include "inode.h"
# include "ialloc.h"
# include "namei.h"
# include "super.h"
# include "truncate.h"
# include "version.h"


static long	_cdecl e_root		(int dev, fcookie *dir);

static long	_cdecl e_lookup		(fcookie *dir, const char *name, fcookie *fc);
static DEVDRV *	_cdecl e_getdev		(fcookie *fc, long *devsp);
static long	_cdecl e_getxattr	(fcookie *fc, XATTR *xattr);

static long	_cdecl e_chattr		(fcookie *fc, int attr);
static long	_cdecl e_chown		(fcookie *fc, int uid, int gid);
static long	_cdecl e_chmod		(fcookie *fc, unsigned mode);

static long	_cdecl e_mkdir		(fcookie *dir, const char *name, unsigned mode);
static long	_cdecl e_rmdir		(fcookie *dir, const char *name);
static long	_cdecl e_creat		(fcookie *dir, const char *name, unsigned mode, int attr, fcookie *fc);
static long	_cdecl e_remove		(fcookie *dir, const char *name);
static long	_cdecl e_getname	(fcookie *root, fcookie *dir, char *pathname, int length);
static long	_cdecl e_rename		(fcookie *olddir, char *oldname, fcookie *newdir, const char *newname);

static long	_cdecl e_opendir	(DIR *dirh, int flag);
static long	_cdecl e_readdir	(DIR *dirh, char *name, int namelen, fcookie *fc);
static long	_cdecl e_rewinddir	(DIR *dirh);
static long	_cdecl e_closedir	(DIR *dirh);

static long	_cdecl e_pathconf	(fcookie *dir, int which);
static long	_cdecl e_dfree		(fcookie *dir, long *buffer);
static long	_cdecl e_wlabel		(fcookie *dir, const char *name);
static long	_cdecl e_rlabel		(fcookie *dir, char *name, int namelen);

static long	_cdecl e_symlink	(fcookie *dir, const char *name, const char *to);
static long	_cdecl e_readlink	(fcookie *fc, char *buf, int len);
static long	_cdecl e_hardlink	(fcookie *fromdir, const char *fromname, fcookie *todir, const char *toname);
static long	_cdecl e_fscntl		(fcookie *dir, const char *name, int cmd, long arg);
static long	_cdecl e_dskchng	(int drv, int mode);

static long	_cdecl e_release	(fcookie *fc);
static long	_cdecl e_dupcookie	(fcookie *dst, fcookie *src);
static long	_cdecl e_sync		(void);


FILESYS ext2_filesys =
{
	NULL,
	/*
	 * FS_KNOPARSE		kernel shouldn't do parsing
	 * FS_CASESENSITIVE	file names are case sensitive
	 * FS_NOXBIT		if a file can be read, it can be executed
	 * FS_LONGPATH		file system understands "size" argument to "getname"
	 * FS_NO_C_CACHE	don't cache cookies for this filesystem
	 * FS_DO_SYNC		file system has a sync function
	 * FS_OWN_MEDIACHANGE	filesystem control self media change (dskchng)
	 */
	FS_CASESENSITIVE | FS_LONGPATH | FS_DO_SYNC | FS_OWN_MEDIACHANGE,
	e_root,
	e_lookup, e_creat, e_getdev, e_getxattr,
	e_chattr, e_chown, e_chmod,
	e_mkdir, e_rmdir, e_remove, e_getname, e_rename,
	e_opendir, e_readdir, e_rewinddir, e_closedir,
	e_pathconf, e_dfree, e_wlabel, e_rlabel,
	e_symlink, e_readlink, e_hardlink, e_fscntl, e_dskchng,
	e_release, e_dupcookie,
	e_sync
};


static long _cdecl
e_root (int drv, fcookie *fc)
{
	SI *s = super [drv];
	
	DEBUG (("Ext2-FS [%c]: e_root enter (s = %lx, mem = %li)", drv+'A', s, memory));
	
	if (!s)
	{
		long i;
		
		i = read_ext2_sb_info (drv);
		if (i)
		{
			DEBUG (("Ext2-FS: e_root leave failure"));
			return i;
		}
		
		s = super [drv];
	}
	
	fc->fs = &ext2_filesys;
	fc->dev = drv;
	fc->aux = 0;
	fc->index = (long) s->root; s->root->links++;
	
	DEBUG (("Ext2-FS: e_root leave ok (mem = %li)", memory));
	return E_OK;
}

static long _cdecl
e_lookup (fcookie *dir, const char *name, fcookie *fc)
{
	COOKIE *c = (COOKIE *) dir->index;
	SI *s = super [dir->dev];
	
	DEBUG (("Ext2-FS [%c]: e_lookup (%s)", dir->dev+'A', name));
	
	*fc = *dir;
	
	/* 1 - itself */
	if (!*name || (name[0] == '.' && name[1] == '\0'))
	{	
		c->links++;
	
		DEBUG (("Ext2-FS: e_lookup: leave ok, (name = \".\")"));
		return E_OK;
	}
	
	/* 2 - parent dir */
	if (name[0] == '.' && name[1] == '.' && name[2] == '\0')
	{
		if (dir->index == (long) s->root)
		{
			DEBUG (("Ext2-FS: e_lookup: leave ok, EMOUNT, (name = \"..\")"));
			return EMOUNT;
		}
	}
	
	{
		_DIR *dentry;
		long ret;
		
		dentry = ext2_search_entry (c, name, _STRLEN (name));
		if (!dentry)
		{
			DEBUG (("Ext2-FS [%c]: e_lookup: leave EFILNF", dir->dev+'A'));
			return EFILNF;
		}
		
		ret = get_cookie (s, dentry->inode, &c);
		if (ret)
		{
			DEBUG (("Ext2-FS: e_lookup: leave ret = %li", ret));
			return ret;
		}
		
		fc->index = (long) c;
	}
	
	DEBUG (("Ext2-FS [%c]: e_lookup: leave ok", dir->dev+'A'));
	return E_OK;
}

static DEVDRV * _cdecl
e_getdev (fcookie *fc, long *devsp)
{
	if (fc->fs == &ext2_filesys)
		return &ext2_dev;
	
	*devsp = EINVFN;
	
	DEBUG (("Ext2-FS: e_getdev: leave failure"));
	return NULL;
}

static long _cdecl
e_getxattr (fcookie *fc, XATTR *xattr)
{
	COOKIE *c = (COOKIE *) fc->index;
	SI *s = super [fc->dev];
	
	register ushort mode = SWAP68_W (c->in.i_mode);
	
	DEBUG (("Ext2-FS [%c]: e_getxattr: #%li", fc->dev+'A', c->inode));
	
	xattr->mode = mode;
	if (EXT2_ISLNK (mode))
	{
		xattr->mode &= ~EXT2_IFLNK;
		xattr->mode |= S_IFLNK;
	}
	if (EXT2_ISFIFO (mode))
	{
		xattr->mode &= ~EXT2_IFIFO;
		xattr->mode |= S_IFIFO;
	}
	
	/* fake attr field a little bit */
	if (EXT2_ISDIR (mode))
		xattr->attr = FA_DIR;
	else
		xattr->attr = (mode & 0222) ? 0 : FA_RDONLY;
	
	xattr->index	= c->inode;
	xattr->dev	= c->dev;
	xattr->rdev 	= c->rdev;
	xattr->nlink	= SWAP68_W (c->in.i_links_count);
	xattr->uid	= SWAP68_W (c->in.i_uid);
	xattr->gid	= SWAP68_W (c->in.i_gid);
	xattr->size 	= SWAP68_L (c->in.i_size);
	xattr->blksize	= s->sbi.s_blocksize;
	xattr->nblocks	= SWAP68_L (c->in.i_blocks);
	*((long *) &(xattr->mtime)) = dostime (SWAP68_L (c->in.i_mtime));
	*((long *) &(xattr->atime)) = dostime (SWAP68_L (c->in.i_atime));
	*((long *) &(xattr->ctime)) = dostime (SWAP68_L (c->in.i_ctime));
	
	DEBUG (("Ext2-FS [%c]: e_getxattr: leave ok", fc->dev+'A'));
	return E_OK;
}

/* the only settable attribute is FA_RDONLY; if the bit is set,
 * the mode is changed so that no write permission exists
 */
static long _cdecl
e_chattr (fcookie *fc, int attr)
{
	COOKIE *c = (COOKIE *) fc->index;
	ushort mode;
	
	DEBUG (("Ext2-FS [%c]: e_chattr", fc->dev+'A'));
	
	if ((c->s->s_flags & MS_RDONLY) || IS_IMMUTABLE (c))
		return EWRPRO;
	
	if ((attr & FA_RDONLY) || (attr == 0))
	{
		mode = SWAP68_W (c->in.i_mode);
		
		if (attr)
		{
			/* turn off write permission */
			mode &= ~(0222);
			
			goto write;
		}
		else
		{
			if ((mode & 0222) == 0)
			{
				/* turn write permission back on */
				mode |= (mode & 0444) >> 1;
				
				goto write;
			}
		}
	}
	
	return E_OK;
	
write:
	c->in.i_mode = SWAP68_W (mode);
	c->in.i_ctime = SWAP68_L (CURRENT_TIME);
	mark_inode_dirty (c);
	
	bio_SYNC_DRV (&bio, c->s->di);
	return E_OK;
}

static long _cdecl
e_chown (fcookie *fc, int uid, int gid)
{
	COOKIE *c = (COOKIE *) fc->index;
	
	DEBUG (("Ext2-FS [%c]: e_chown", fc->dev+'A'));
	
	if ((c->s->s_flags & MS_RDONLY) || IS_IMMUTABLE (c))
		return EWRPRO;
	
	if (uid != -1) c->in.i_uid = SWAP68_W (uid);
	if (gid != -1) c->in.i_gid = SWAP68_W (gid);
	
	c->in.i_ctime = SWAP68_L (CURRENT_TIME);
	
	mark_inode_dirty (c);
	
	bio_SYNC_DRV (&bio, c->s->di);
	return E_OK;
}

static long _cdecl
e_chmod (fcookie *fc, unsigned mode)
{
	COOKIE *c = (COOKIE *) fc->index;
	
	DEBUG (("Ext2-FS [%c]: e_chmod", fc->dev+'A'));
	
	if ((c->s->s_flags & MS_RDONLY) || IS_IMMUTABLE (c))
		return EWRPRO;
	
	c->in.i_mode = SWAP68_W ((SWAP68_W (c->in.i_mode) & S_IFMT) | (mode & S_IALLUGO));
	c->in.i_ctime = SWAP68_L (CURRENT_TIME);
	
	mark_inode_dirty (c);
	
	bio_SYNC_DRV (&bio, c->s->di);
	return E_OK;
}

static long _cdecl
e_mkdir (fcookie *dir, const char *name, unsigned mode)
{
	SI *s = super [dir->dev];
	COOKIE *dirc = (COOKIE *) dir->index;
	COOKIE *inode;
	
	long namelen = _STRLEN (name);
	
	UNIT *u;
	UNIT *dir_block;
	ext2_d2 *de;
	
	long err = E_OK;
	
	
	DEBUG (("Ext2-FS [%c]: e_mkdir", dir->dev+'A'));
	
	if (namelen > EXT2_NAME_LEN)
		return ENAMETOOLONG;
	
	if ((s->s_flags & MS_RDONLY) || IS_IMMUTABLE (dirc))
		return EWRPRO;
	
	if (SWAP68_W (dirc->in.i_links_count) >= EXT2_LINK_MAX)
		return EMLINK;
	
	inode = ext2_new_inode (dirc, EXT2_IFDIR, &err);
	if (!inode)
		return EIO;
	
	inode->in.i_size = SWAP68_L (s->sbi.s_blocksize);
	inode->in.i_blocks = 0;	
	
	dir_block = ext2_bread (inode, 0, &err);
	if (!dir_block)
	{
		/* delete inode */
		err = EIO;
		goto out_no_entry;
	}
	
	de = (ext2_d2 *) dir_block->data;
	de->inode = SWAP68_L (inode->inode);
	de->name_len = 1;
	de->rec_len = SWAP68_W (EXT2_DIR_REC_LEN (de->name_len));
	strcpy (de->name, ".");
	
	if (EXT2_HAS_INCOMPAT_FEATURE (s, EXT2_FEATURE_INCOMPAT_FILETYPE))
		de->file_type = EXT2_FT_DIR;
	
	de = (ext2_d2 *) ((char *) de + SWAP68_W (de->rec_len));
	de->inode = SWAP68_L (dirc->inode);
	de->rec_len = SWAP68_W (s->sbi.s_blocksize - EXT2_DIR_REC_LEN (1));
	de->name_len = 2;
	strcpy (de->name, "..");
	
	if (EXT2_HAS_INCOMPAT_FEATURE (s, EXT2_FEATURE_INCOMPAT_FILETYPE))
		de->file_type = EXT2_FT_DIR;
	
	inode->in.i_links_count = SWAP68_W (2);
	bio_MARK_MODIFIED (&bio, dir_block);
	
	inode->in.i_mode = S_IFDIR | (mode & (S_IRWXUGO | S_ISVTX) /*& ~current->fs->umask*/);
	if (SWAP68_W (dirc->in.i_mode) & S_ISGID)
		inode->in.i_mode |= S_ISGID;
	inode->in.i_mode = SWAP68_W (inode->in.i_mode);
	mark_inode_dirty (inode);
	
	u = ext2_add_entry (dirc, name, namelen, &de, &err);
	if (!u)
		goto out_no_entry;
	
	de->inode = SWAP68_L (inode->inode);
	if (EXT2_HAS_INCOMPAT_FEATURE (s, EXT2_FEATURE_INCOMPAT_FILETYPE))
		de->file_type = EXT2_FT_DIR;
	bio_MARK_MODIFIED (&bio, u);
	
	dirc->in.i_version = SWAP68_L (++event);
	dirc->in.i_links_count = SWAP68_W (SWAP68_W (dirc->in.i_links_count) + 1);
	dirc->in.i_flags = SWAP68_L (SWAP68_L (dirc->in.i_flags) & ~EXT2_BTREE_FL);
	mark_inode_dirty (dirc);
	
	{
		_DIR *dentry;
		
		dentry = d_get_dir (dirc, inode->inode, name, namelen);
		if (dentry)
			dentry->links--;
	}
	
out:
	/* release the cookie */
	rel_cookie (inode);
	
	bio_SYNC_DRV (&bio, s->di);
	
	DEBUG (("Ext2-FS [%c]: e_mkdir: leave (%li)", err));
	return err;
	
out_no_entry:
	inode->in.i_links_count = 0;
	mark_inode_dirty (inode);
	goto out;
}

/* routine to check that the specified directory is empty (for rmdir)
 */
static long
empty_dir (COOKIE *inode)
{
	SI *s = inode->s;
	UNIT *u;
	ext2_d2 *de, *de1;
	ulong offset;
	long err;
	
	if (SWAP68_L (inode->in.i_size) < EXT2_DIR_REC_LEN (1) + EXT2_DIR_REC_LEN (2)
		|| !(u = ext2_read (inode, 0, &err)))
	{
	    	ALERT (("Ext2-FS [%c]: empty_dir: bad directory (dir #%li) - no data block", 'A'+inode->dev, inode->inode));
		return 1;
	}
	
	de = (ext2_d2 *) u->data;
	de1 = (ext2_d2 *) ((char *) de + SWAP68_W (de->rec_len));
	
	if (SWAP68_L (de->inode) != inode->inode
		|| !SWAP68_L (de1->inode)
		|| strcmp (".", de->name)
		|| strcmp ("..", de1->name))
	{
	    	ALERT (("Ext2-FS [%c]: empty_dir: bad directory (dir #%li) - no `.' or `..'", 'A'+inode->dev, inode->inode));
		return 1;
	}
	
	offset = SWAP68_W (de->rec_len) + SWAP68_W (de1->rec_len);
	de = (ext2_d2 *) ((char *) de1 + SWAP68_W (de1->rec_len));
	while (offset < SWAP68_L (inode->in.i_size))
	{
		if (!u || (void *) de >= (void *) (u->data + s->sbi.s_blocksize))
		{
			u = ext2_bread (inode, offset >> EXT2_BLOCK_SIZE_BITS (s), &err);
			if (!u)
			{
				ALERT (("Ext2-FS [%c]: empty_dir: directory #%lu contains a hole at offset %lu",
					inode->inode, offset));
				
				offset += s->sbi.s_blocksize;
				continue;
			}
			
			de = (ext2_d2 *) u->data;
		}
		
		if (!ext2_check_dir_entry ("empty_dir", inode, de, u, offset))
			return 1;
		
		if (SWAP68_L (de->inode))
			return 0;
		
		offset += SWAP68_W (de->rec_len);
		de = (ext2_d2 *) ((char *) de + SWAP68_W (de->rec_len));
	}
	
	return 1;
}

static long _cdecl
e_rmdir (fcookie *dir, const char *name)
{
	SI *s = super [dir->dev];
	COOKIE *dirc = (COOKIE *) dir->index;
	COOKIE *inode;
	_DIR *dentry;
	
	long namelen = _STRLEN (name);
	long retval = E_OK;
	
	
	DEBUG (("Ext2-FS [%c]: e_rmdir (%s)", dir->dev+'A', name));
	
	if (namelen > EXT2_NAME_LEN)
		return ENAMETOOLONG;
	
	if ((s->s_flags & MS_RDONLY) || IS_IMMUTABLE (dirc))
		return EWRPRO;
	
	dentry = ext2_search_entry (dirc, name, namelen);
	if (!dentry)
		return ENOENT;
	
	retval = get_cookie (s, dentry->inode, &inode);
	if (retval)
		return retval;
	
	if (IS_IMMUTABLE (inode))
	{
		retval = EWRPRO;
		goto out;
	}
	
# if 0
	if ((SWAP68_W (dirc->in.i_mode) & S_ISVTX)
		&& current->fsuid != SWAP68_W (inode->in.i_uid)
		&& current->fsuid != SWAP68_W (dirc->in.i_uid)
		&& !capable (CAP_FOWNER))
	{
		retval = EACCDN;
		goto out;
	}
# endif
	
	/* we may not delete ".", but "../dir" is ok */
	if (inode == dirc)
	{
		retval = EACCDN;
		goto out;
	}
	
	if (!S_ISDIR (SWAP68_W (inode->in.i_mode)))
	{
		retval = ENOTDIR;
		goto out;
	}
	
	if ((inode->dev != dir->dev) || (inode->inode != dentry->inode))
	{
		retval = EIO;
		goto out;
	}
	
	if (!empty_dir (inode))
	{
		retval = ENOTEMPTY;
	}
	else
	{
		ext2_d2 *de;
		UNIT *u;
		
		u = ext2_find_entry (dirc, name, namelen, &de);
		if (!u)
		{
			retval = ENOENT;
			goto out;
		}
		
		if (SWAP68_L (de->inode) != inode->inode)
		{
			retval = ENOENT;
			goto out;
		}
		
		retval = ext2_delete_entry (de, u);
		dirc->in.i_version = SWAP68_L (++event);
	}
	
	if (retval)
		goto out;
	
	if (SWAP68_W (inode->in.i_links_count) != 2)
		ALERT (("Ext2-FS [%c]: e_rmdir: empty directory has nlink != 2 (%ld)", SWAP68_W (inode->in.i_links_count)));
	
	inode->in.i_version = SWAP68_L (++event);
	inode->in.i_links_count = 0;
	inode->in.i_ctime = SWAP68_L (CURRENT_TIME);
	mark_inode_dirty (inode);
	
	dirc->in.i_links_count = SWAP68_W (SWAP68_W (dirc->in.i_links_count) - 1);
	dirc->in.i_ctime = dirc->in.i_mtime = inode->in.i_ctime;
	dirc->in.i_flags = SWAP68_L (SWAP68_L (dirc->in.i_flags) & ~EXT2_BTREE_FL);
	mark_inode_dirty (dirc);
	
	/* update directory cache */
	d_del_dir (dentry);
	
out:
	/* release the cookie */
	rel_cookie (inode);
	
	bio_SYNC_DRV (&bio, s->di);
	
	DEBUG (("Ext2-FS [%c]: e_rmdir: leave (%li)", dir->dev+'A', retval));
	return retval;
}

static long _cdecl
e_creat (fcookie *dir, const char *name, unsigned mode, int attr, fcookie *fc)
{
	COOKIE *dirc = (COOKIE *) dir->index;
	COOKIE *inode;
	UNIT *u;
	ext2_d2 *de;
	
	long err = EIO;
	
	
	DEBUG (("Ext2-FS [%c]: e_creat enter (%s)", dir->dev+'A', name));
	
	if ((dirc->s->s_flags & MS_RDONLY) || IS_IMMUTABLE (dirc))
		return EWRPRO;
	
	inode = ext2_new_inode (dirc, mode, &err);
	if (!inode)
		return err;
	
	u = ext2_add_entry (dirc, name, strlen (name), &de, &err);
	if (!u)
	{
		inode->in.i_links_count = SWAP68_W (SWAP68_W (inode->in.i_links_count) - 1);
		mark_inode_dirty (inode);
		
		/* free the inode */
		ext2_delete_inode (inode);
		
		return err;
	}
	de->inode = SWAP68_L (inode->inode);
	
	if (EXT2_HAS_INCOMPAT_FEATURE (dirc->s, EXT2_FEATURE_INCOMPAT_FILETYPE))
		de->file_type = EXT2_FT_REG_FILE;
	
	bio_MARK_MODIFIED (&bio, u);
	
	dirc->in.i_version = SWAP68_L (++event);
	mark_inode_dirty (dirc);
	
	{
		_DIR *dentry;
		
		dentry = d_get_dir (dirc, inode->inode, name, strlen (name));
		if (dentry)
			dentry->links--;
	}
	
	*fc = *dir;
	fc->index = (long) inode;
	
	bio_SYNC_DRV (&bio, inode->s->di);
	
	DEBUG (("Ext2-FS [%c]: e_creat leave OK (#%li)", dir->dev+'A', inode->inode));
	return E_OK;
}

static long _cdecl
e_remove (fcookie *dir, const char *name)
{
	COOKIE *dirc = (COOKIE *) dir->index;
	COOKIE *inode;
	UNIT *u;
	ext2_d2 *de;
	_DIR *dentry;
	long namelen = strlen (name);
	long retval;
	
	
	DEBUG (("Ext2-FS [%c]: e_remove: enter (%s)", dir->dev+'A', name));
	
	if (namelen > EXT2_NAME_LEN)
		return ENAMETOOLONG;
	
	if ((dirc->s->s_flags & MS_RDONLY) || IS_IMMUTABLE (dirc))
		return EWRPRO;
	
	dentry = ext2_search_entry (dirc, name, namelen);
	if (!dentry)
		return ENOENT;
	
	retval = get_cookie (dirc->s, dentry->inode, &inode);
	if (retval)
		return retval;
	
	if (EXT2_ISDIR (SWAP68_W (inode->in.i_mode)))
	{
		retval = EACCDN;
		goto out;
	}
	
	if (IS_APPEND (inode) || IS_IMMUTABLE (inode))
	{
		retval = EACCDN;
		goto out;
	}
	
# if 0
	if ((SWAP68_W (dir->in.i_mode) & S_ISVTX)
		&& current->fsuid != SWAP68_W (inode->in.i_uid)
		&& current->fsuid != SWAP68_W (dirc->in.i_uid)
		&& !capable (CAP_FOWNER)*/)
	{
		retval = EACCDN;
		goto out;
	}
# endif
	
	if (dentry->inode != inode->inode)
	{
		retval = EIO;
		goto out;
	}
	
	if (!inode->in.i_links_count)
	{
		ALERT (("Ext2_FS [%c]: ext2_unlink: Deleting nonexistent file (%lu), %d", inode->dev+'A', inode->inode, SWAP68_W (inode->in.i_links_count)));
		inode->in.i_links_count = SWAP68_W (1);
	}
	
	u = ext2_find_entry (dirc, name, namelen, &de);
	if (!u)
	{
		retval = ENOENT;
		goto out;
	}
	
	retval = ext2_delete_entry (de, u);
	if (retval)
	{
		retval = EACCDN;
		goto out;
	}
	
	bio_MARK_MODIFIED (&bio, u);
	
	dirc->in.i_version = SWAP68_L (++event);
	dirc->in.i_ctime = dirc->in.i_mtime = SWAP68_L (CURRENT_TIME);
	dirc->in.i_flags = SWAP68_L (SWAP68_L (dirc->in.i_flags) & ~EXT2_BTREE_FL);
	mark_inode_dirty (dirc);
	
	inode->in.i_links_count = SWAP68_W (SWAP68_W (inode->in.i_links_count) - 1);
	inode->in.i_ctime = dirc->in.i_ctime;
	mark_inode_dirty (inode);
	
	/* update directory cache */
	d_del_dir (dentry);
	
	retval = E_OK;
	
out:
	/* release cookie (also delete inode if neccessary) */
	rel_cookie (inode);
	
	bio_SYNC_DRV (&bio, dirc->s->di);
	
	DEBUG (("Ext2-FS [%c]: e_remove: leave (%li)", dir->dev+'A', retval));
	return retval;
}

static long _cdecl
e_getname (fcookie *root, fcookie *dir, char *pathname, int length)
{
	SI *s = super [dir->dev];
	ulong inum = ((COOKIE *) dir->index)->inode;
	
	char *dst = pathname;
	long len = 0;
	
	DEBUG (("Ext2-FS [%c]: e_getname: #%li -> #%li", root->dev+'A', ((COOKIE *) root->index)->inode, ((COOKIE *) dir->index)->inode));
	ASSERT ((((COOKIE *) root->index)->inode == EXT2_ROOT_INO));
	
	*pathname = '\0';
	length--;
	
	while (inum != EXT2_ROOT_INO)
	{
		COOKIE *c;
		_DIR *dentry;
		ulong pinum;
		ulong size;
		long block;
		long offset;
		long r;
		
		r = get_cookie (s, inum, &c);
		if (r)
		{
			DEBUG (("Ext2-FS: e_getname: get_cookie (#%li) fail: r = %li", inum, r));
			return r;
		}
		
		dentry = ext2_search_entry (c, "..", 2);
		if (!dentry)
		{
			/* If this happens we're in trouble */
			
			ALERT (("Ext2-FS [%c]: e_getname: no '..' in inode #%li", c->dev+'A', c->inode));
			return inum;
		}
		c->links--;
		
		pinum = dentry->inode;
		
		r = get_cookie (s, pinum, &c);
		if (r)
		{
			DEBUG (("Ext2-FS: e_getname: get_cookie (#%li) fail: r = %li", inum, r));
			return r;
		}
		
		r = 1;
		size = SWAP68_L (c->in.i_size);
		for (block = 0, offset = 0; offset < size; block++)
		{
			UNIT *u;
			ext2_d2 *de;
			char *upper;
			
			u = ext2_read (c, block, &r);
			if (!u)
			{
				DEBUG (("Ext2-FS: e_getname: ext2_read fail (%li)", block));
				
				r = EINTRN;
				break;
			}
			
			de = (ext2_d2 *) u->data;
			
			upper = (char *) de + s->sbi.s_blocksize;
			while ((char *) de < upper)
			{
				register long de_len;
				
				if (inum == SWAP68_L (de->inode))
				{
					register long i = de->name_len;
					register char *src;
					
					len += i;
					if (len >= length)
					{
						DEBUG (("Ext2-FS: e_getname: ENAMETOOLONG"));
						
						r = ENAMETOOLONG;
						break;
					}
					
					src = de->name + i - 1;
					while (i--)
					{
						*dst++ = *src--;
					}
					
					*dst++ = '\\';
					*dst = '\0';
					
					inum = pinum;
					
					r = 0;
					break;
				}
				
				de_len = SWAP68_W (de->rec_len);
				if (de_len <= 0)
				{
					DEBUG (("Ext2-FS: e_getname: inum not found"));
					
					r = EINTRN;
					break;
				}
				
				offset += de_len;
				de = (ext2_d2 *) ((char *) de + de_len);
			}
			
			if (r <= 0)
				break;
		}
		
		c->links--;
		
		if (r)
		{
			DEBUG (("Ext2-FS: e_getname: leave failure (r = %li)", r));
			return r;
		}
	}
	
	strrev (pathname);
	
	DEBUG (("Ext2-FS: e_getname: leave ok (%s)", pathname));
	return E_OK;
}

static long _cdecl
e_rename (fcookie *olddir, char *oldname, fcookie *newdir, const char *newname)
{
# define PARENT_INO(buffer) \
	(((ext2_d2 *) ((char *) (buffer) + SWAP68_W (((ext2_d2 *) (buffer))->rec_len)))->inode)
	
	SI *s = super [olddir->dev];
	
	COOKIE *olddirc = (COOKIE *) olddir->index;
	COOKIE *newdirc = (COOKIE *) newdir->index;
	COOKIE *inode;
	
	_DIR *olddentry;
	_DIR *tmpdentry;
	
	UNIT *old_u = NULL;
	UNIT *new_u = NULL;
	UNIT *dir_u = NULL;
	
	ext2_d2 *old_de;
	ext2_d2 *new_de;
	
	long retval = ENOENT;
	
	
	DEBUG (("Ext2-FS [%c]: e_rename: #%li: %s -> #%li: %s", olddir->dev+'A', olddirc->inode, oldname, newdirc->inode, newname));
	
	if ((s->s_flags & MS_RDONLY)
		|| IS_IMMUTABLE (olddirc)
		|| IS_IMMUTABLE (newdirc))
	{
		return EWRPRO;
	}
	
	if (!EXT2_ISDIR (SWAP68_W (olddirc->in.i_mode))
		|| !EXT2_ISDIR (SWAP68_W (newdirc->in.i_mode)))
	{
		return EACCDN;
	}
	
	olddentry = ext2_search_entry (olddirc, oldname, strlen (oldname));
	if (!olddentry)
		goto end_rename;
	
	old_u = ext2_find_entry (olddirc, oldname, strlen (oldname), &old_de);
	if (!old_u)
		goto end_rename;
	
	bio.lock (old_u);
	
	DEBUG (("olddentry->inode = %li <-> old_de->inode = %li", olddentry->inode, SWAP68_L (old_de->inode)));
	retval = get_cookie (s, olddentry->inode, &inode);
	if (retval)
		goto end_rename;
	
	retval = EACCDN;
# if 0
	if ((old_dir->i_mode & S_ISVTX)
		&& current->fsuid != old_inode->i_uid
		&& current->fsuid != old_dir->i_uid && !capable (CAP_FOWNER))
		goto end_rename;
# endif
	if (IS_APPEND (inode) || IS_IMMUTABLE (inode))
		goto end_rename;
	
	if (EXT2_ISDIR (SWAP68_W (inode->in.i_mode)))
	{
		if (SWAP68_W (newdirc->in.i_links_count) >= EXT2_LINK_MAX)
		{
			retval = EMLINK;
			goto end_rename;
		}
		
		dir_u = ext2_read (inode, 0, &retval);
		if (!dir_u)
			goto end_rename;
		
		bio.lock (dir_u);
		
		if (SWAP68_L (PARENT_INO (dir_u->data)) != olddirc->inode)
			goto end_rename;
	}
	
	tmpdentry = ext2_search_entry (newdirc, newname, strlen (newname));
	if (tmpdentry)
		goto end_rename;
	
	new_u = ext2_add_entry (newdirc, newname, strlen (newname), &new_de, &retval);
	if (!new_u)
		goto end_rename;
	
	bio.lock (new_u);
	
	new_de->inode = SWAP68_L (inode->inode);
	if (EXT2_HAS_INCOMPAT_FEATURE (s, EXT2_FEATURE_INCOMPAT_FILETYPE))
		new_de->file_type = old_de->file_type;
	
	bio_MARK_MODIFIED (&bio, new_u);
	bio.unlock (new_u); new_u = NULL;
	
	ext2_delete_entry (old_de, old_u);
	bio.unlock (old_u); old_u = NULL;
	
	if (newdirc != olddirc)
	{
		newdirc->in.i_version = SWAP68_L (++event);
		newdirc->in.i_ctime = newdirc->in.i_mtime = SWAP68_L (CURRENT_TIME);
		newdirc->in.i_flags = SWAP68_L (SWAP68_L (newdirc->in.i_flags) & ~EXT2_BTREE_FL);
		
		if (EXT2_ISDIR (SWAP68_W (inode->in.i_mode)))
		{
			olddirc->in.i_links_count = SWAP68_W (SWAP68_W (olddirc->in.i_links_count) - 1);
			newdirc->in.i_links_count = SWAP68_W (SWAP68_W (newdirc->in.i_links_count) + 1);
			
			PARENT_INO (dir_u->data) = SWAP68_L (newdirc->inode);
			
			bio_MARK_MODIFIED (&bio, dir_u);
			bio.unlock (dir_u); dir_u = NULL;
		}
		
		mark_inode_dirty (newdirc);
	}
	
	olddirc->in.i_version = SWAP68_L (++event);
	olddirc->in.i_ctime = olddirc->in.i_mtime = SWAP68_L (CURRENT_TIME);
	olddirc->in.i_flags = SWAP68_L (SWAP68_L (olddirc->in.i_flags) & ~EXT2_BTREE_FL);
	mark_inode_dirty (olddirc);
	
	d_del_dir (olddentry);
	d_get_dir (newdirc, inode->inode, newname, strlen (newname));
	
	retval = E_OK;
	
end_rename:
	
	if (old_u) bio.unlock (old_u);
	if (new_u) bio.unlock (new_u);
	if (dir_u) bio.unlock (dir_u);
	
	DEBUG (("Ext2-FS [%c]: e_rename: leave r = %li", olddir->dev+'A', retval));
	return retval;
}


# define DIR_BLOCKS(d)	(*(long *) &(d)->fsstuff [0])
# define DIR_BLOCK(d)	(*(long *) &(d)->fsstuff [4])
# define DIR_OFFSET(d)	(*(long *) &(d)->fsstuff [8])

static long _cdecl
e_opendir (DIR *dirh, int flag)
{
	COOKIE *c = (COOKIE *) dirh->fc.index;
	SI *s = super [dirh->fc.dev];
	
	DEBUG (("Ext2-FS [%c]: e_opendir: #%li", dirh->fc.dev+'A', c->inode));
	
	c->links++;
	
	DIR_BLOCKS (dirh) = SWAP68_L (c->in.i_size) >> EXT2_BLOCK_SIZE_BITS (s);
	DIR_BLOCK (dirh) = 0;
	DIR_OFFSET (dirh) = 0;
	
	DEBUG (("Ext2-FS [%c]: e_opendir: leave ok", dirh->fc.dev+'A'));
	return E_OK;
}

static long _cdecl
e_readdir (DIR *dirh, char *name, int namelen, fcookie *fc)
{
	COOKIE *c = (COOKIE *) dirh->fc.index;
	SI *s = super [dirh->fc.dev];
	ext2_d2 *de;
	
	DEBUG (("Ext2-FS [%c]: e_readdir: #%li", dirh->fc.dev+'A', c->inode));
	
	do {
		UNIT *u;
		
		if (DIR_BLOCK (dirh) >= DIR_BLOCKS (dirh))
			return ENMFIL;
		
		u = ext2_read (c, DIR_BLOCK (dirh), NULL);
		if (!u)
			return ERROR;
		
		de = (ext2_d2 *) (u->data + DIR_OFFSET (dirh));
		
		DIR_OFFSET (dirh) += SWAP68_W (de->rec_len);
		DIR_BLOCK  (dirh) += DIR_OFFSET (dirh) >> EXT2_BLOCK_SIZE_BITS (s);
		DIR_OFFSET (dirh) &= EXT2_BLOCK_SIZE_MASK (s);
	}
	while (de->inode == 0);
	
	{
		COOKIE *new;
		long r;
		
		r = get_cookie (s, SWAP68_L (de->inode), &new);
		if (r)
		{
			DEBUG (("Ext2-FS: e_readdir: get_cookie fail (%li)", r));
			return r;
		}
		
		/* setup file cookie */
		fc->fs = &ext2_filesys;
		fc->dev = dirh->fc.dev;
		fc->aux = 0;
		fc->index = (long) new;
		
		if ((dirh->flags & TOS_SEARCH) == 0)
		{
			*(long *) name = new->inode;
			namelen -= 4;
			name += 4;
		}
	}
	
	_STRNCPY (name, de->name, MIN (namelen, de->name_len));
	
	if (namelen <= de->name_len)
		return ENAMETOOLONG;
	
	name [de->name_len] = '\0';
	
	DEBUG (("Ext2-FS [%c]: e_readdir ok (%s)", dirh->fc.dev+'A', name));
	return E_OK;
}

static long _cdecl
e_rewinddir (DIR *dirh)
{
	DEBUG (("Ext2-FS [%c]: e_rewinddir: #%li", dirh->fc.dev+'A', ((COOKIE *) dirh->fc.index)->inode));
	
	DIR_BLOCK (dirh)  = 0;
	DIR_OFFSET (dirh) = 0;
	
	return E_OK;
}

static long _cdecl
e_closedir (DIR *dirh)
{
	COOKIE *c = (COOKIE *) dirh->fc.index;
	
	DEBUG (("Ext2-FS [%c]: e_closedir: #%li", dirh->fc.dev+'A', c->inode));
	
	rel_cookie (c);
	return E_OK;
}

static long _cdecl
e_pathconf (fcookie *dir, int which)
{
	DEBUG (("Ext2-FS [%c]: e_pathconf (%i)", dir->dev+'A', which));
	
	switch (which)
	{
		case DP_INQUIRE:	return DP_VOLNAMEMAX;
		case DP_IOPEN:		return UNLIMITED;
		case DP_MAXLINKS:	return EXT2_LINK_MAX;
		case DP_PATHMAX:	return UNLIMITED;
		case DP_NAMEMAX:	return EXT2_NAME_LEN;
		case DP_ATOMIC:		return (super [dir->dev])->sbi.s_blocksize;
		case DP_TRUNC:		return DP_NOTRUNC;
		case DP_CASE:		return DP_CASESENS;
		case DP_MODEATTR:	return (DP_ATTRBITS
						| DP_MODEBITS
						| DP_FT_DIR
						| DP_FT_REG
						| DP_FT_LNK
					);
		case DP_XATTRFIELDS:	return (DP_INDEX
						| DP_DEV
						| DP_NLINK
						| DP_UID
						| DP_GID
						| DP_BLKSIZE
						| DP_SIZE
						| DP_NBLOCKS
						| DP_ATIME
						| DP_CTIME
						| DP_MTIME
					);
		case DP_VOLNAMEMAX:	return 0;
	}
	
	DEBUG (("Ext2-FS: e_pathconf: leave failure"));
	return EINVFN;
}

static long _cdecl
e_dfree (fcookie *dir, long *buffer)
{
	SI *s = super [dir->dev];
	
	DEBUG (("Ext2-FS [%c]: e_dfree", dir->dev+'A'));
	
	buffer[0] = SWAP68_L (s->sbi.s_sb->s_free_blocks_count);
	buffer[1] = s->sbi.s_blocks_count;
	buffer[2] = s->sbi.s_blocksize;
	buffer[3] = 1;
	
	return E_OK;
}

static long _cdecl
e_wlabel (fcookie *dir, const char *name)
{
	long namelen = strlen (name);
	long r = E_OK;
	
	DEBUG (("Ext2-FS [%c]: e_wlabel enter (%s)", dir->dev+'A', name));
	
	if (namelen > 16)
	{
		r = ENAMETOOLONG;
	}
	else
	{
		SI *s = super [dir->dev];
		
		if (s->s_flags & MS_RDONLY)
		{
			r = EWRPRO;
		}
		else
		{
			strncpy (s->sbi.s_sb->s_volume_name, name, 16);
			bio_MARK_MODIFIED (&bio, s->sbi.s_sb_unit);
			
			bio_SYNC_DRV (&bio, s->di);
		}
	}
	
	DEBUG (("Ext2-FS [%c]: e_wlabel leave (ret = %li)", dir->dev+'A', r));
	return r;
}

static long _cdecl
e_rlabel (fcookie *dir, char *name, int namelen)
{
	SI *s = super [dir->dev];
	long ret = E_OK;
	
	long len;
	char *src;
	
	DEBUG (("Ext2-FS [%c]: e_rlabel enter", dir->dev+'A'));
	
	len = 0;
	src = s->sbi.s_sb->s_volume_name;
	while (len++ < 16 && *src++)
		;
	
	if (namelen < len)
	{
		len = namelen;
		ret = ENAMETOOLONG;
	}
	
	strncpy (name, s->sbi.s_sb->s_volume_name, len);
	name [len] = '\0';
	
	DEBUG (("Ext2-FS [%c]: e_rlabel leave (ret = %li)", dir->dev+'A', ret));
	return ret;
}

static long _cdecl
e_symlink (fcookie *dir, const char *name, const char *to)
{
	COOKIE *dirc = (COOKIE *) dir->index;
	COOKIE *inode;
	UNIT *u = NULL;
	char *link;
	long tolen;
	long err;
	
	DEBUG (("Ext2-FS [%c]: e_symlink: enter (%s, %s)", dir->dev+'A', name, to));
	
	if ((dirc->s->s_flags & MS_RDONLY) || IS_IMMUTABLE (dirc))
		return EWRPRO;
	
	tolen = strlen (to);
	if (tolen > dirc->s->sbi.s_blocksize)
		return EACCDN;
	
	inode = ext2_new_inode (dirc, EXT2_IFLNK, &err);
	if (!inode)
		return err;
	
	inode->in.i_mode = SWAP68_W (EXT2_IFLNK | S_IRWXUGO);
	
	if (tolen >= sizeof (inode->in.i_block))
	{
		DEBUG (("tolen = %ld, normal symlink", tolen));
		
		u = ext2_bread (inode, 0, &err);
		if (!u)
			goto out_no_entry;
		
		link = u->data;
	}
	else
	{
		DEBUG (("tolen = %ld, fast symlink", tolen));
		
		link = (char *) inode->in.i_block;
	}
	
	{
		long i = 0;
		char c;
		
		while (i < inode->s->sbi.s_blocksize - 1 && (c = *to++))
			link [i++] = c;
		
		link [i] = 0;
		
		inode->in.i_size = SWAP68_L (i);
		mark_inode_dirty (inode);
		
		if (u)
			bio_MARK_MODIFIED (&bio, u);
	}
	
	{
		ext2_d2 *de;
		
		u = ext2_add_entry (dirc, name, strlen (name), &de, &err);
		if (!u)
		{
			err = EACCDN;
			goto out_no_entry;
		}
		
		de->inode = SWAP68_L (inode->inode);
		
		if (EXT2_HAS_INCOMPAT_FEATURE (dirc->s, EXT2_FEATURE_INCOMPAT_FILETYPE))
			de->file_type = EXT2_FT_SYMLINK;
		
		bio_MARK_MODIFIED (&bio, u);
	}
	
	dirc->in.i_version = SWAP68_L (++event);
	mark_inode_dirty (dirc);
	
	{
		_DIR *dentry;
		
		dentry = d_get_dir (dirc, inode->inode, name, strlen (name));
		if (dentry)
			dentry->links--;
	}
	
	err = E_OK;
out:
	rel_cookie (inode);
	
	bio_SYNC_DRV (&bio, dirc->s->di);
	
	DEBUG (("Ext2-FS [%c]: e_symlink: leave (%li)", dir->dev+'A', err));
	return err;
	
out_no_entry:
	inode->in.i_links_count = SWAP68_W (SWAP68_W (inode->in.i_links_count) - 1);
	mark_inode_dirty (inode);
	goto out;
}

static long _cdecl
e_readlink (fcookie *fc, char *buf, int len)
{
	COOKIE *inode = (COOKIE *) fc->index;
	UNIT *u = NULL;
	char *link;
	long i;

	if (len > inode->s->sbi.s_blocksize - 1)
		len = inode->s->sbi.s_blocksize - 1;
	
	if (inode->in.i_blocks)
	{
		long err;
		
		u = ext2_read (inode, 0, &err);
		if (!u)
			return err;
		
		link = u->data;
	}
	else
	{
		link = (char *) inode->in.i_block;
	}
	
	i = strlen (link);	ASSERT ((i == SWAP68_L (inode->in.i_size)));
	strncpy (buf, link, MIN (i, len));
	
	if (i > len)
	{
		buf [len] = '\0';
		return ENAMETOOLONG;
	}
	
	buf [i] = '\0';
	return E_OK;
}

long
e_hardlink (fcookie *fromdir, const char *fromname, fcookie *todir, const char *toname)
{
	COOKIE *fromdirc = (COOKIE *) fromdir->index;
	COOKIE *todirc = (COOKIE *) todir->index;
	COOKIE *inode;
	_DIR *dentry;
	ext2_d2 *de;
	UNIT *u;
	long err;
	
	ushort i_mode;
	
	
	DEBUG (("Ext2-FS [%c]: e_hardlink enter (%s -> %s)", fromdir->dev+'A', fromname, toname));
	
	if ((fromdirc->s->s_flags & MS_RDONLY) || IS_IMMUTABLE (todirc))
		return EWRPRO;
	
	dentry = ext2_search_entry (fromdirc, fromname,  strlen (fromname));
	if (!dentry)
		return EACCDN;
	
	err = get_cookie (fromdirc->s, dentry->inode, &inode);
	if (err)
		return err;
	
	i_mode = SWAP68_W (inode->in.i_mode);
	
	if (EXT2_ISDIR (i_mode))
	{
		err = EACCDN;
		goto out;
	}
	
	if (IS_APPEND (inode) || IS_IMMUTABLE (inode))
	{
		err =  EPERM;
		goto out;
	}
	
	if (SWAP68_W (inode->in.i_links_count) >= EXT2_LINK_MAX)
	{
		err =  EMLINK;
		goto out;
	}
	
	u = ext2_add_entry (todirc, toname, strlen (toname), &de, &err);
	if (!u)
		goto out;
	
	de->inode = SWAP68_L (inode->inode);
	if (EXT2_HAS_INCOMPAT_FEATURE (inode->s, EXT2_FEATURE_INCOMPAT_FILETYPE))
	{
		if (EXT2_ISREG (i_mode))
			de->file_type = EXT2_FT_REG_FILE;
		else if (EXT2_ISDIR (i_mode))
			de->file_type = EXT2_FT_DIR;
		else if (EXT2_ISLNK (i_mode))
			de->file_type = EXT2_FT_SYMLINK;
		else if (EXT2_ISCHR (i_mode))
			de->file_type = EXT2_FT_CHRDEV;
		else if (EXT2_ISBLK (i_mode))
			de->file_type = EXT2_FT_BLKDEV;
		else if (EXT2_ISFIFO (i_mode))  
			de->file_type = EXT2_FT_FIFO;
	}
	bio_MARK_MODIFIED (&bio, u);
	
	todirc->in.i_version = SWAP68_L (++event);
	mark_inode_dirty (todirc);
	
	inode->in.i_links_count = SWAP68_W (SWAP68_W (inode->in.i_links_count) + 1);
	inode->in.i_ctime = SWAP68_L (CURRENT_TIME);
	mark_inode_dirty (inode);
	
	err = E_OK;
	
out:
	/* release cookie */
	rel_cookie (inode);
	
	bio_SYNC_DRV (&bio, fromdirc->s->di);
	
	DEBUG (("Ext2-FS [%c]: e_hardlink: leave (%li)", fromdir->dev+'A', err));
	return err;
}

static long _cdecl
e_fscntl (fcookie *dir, const char *name, int cmd, long arg)
{
	SI *s = super [dir->dev];
	
	DEBUG (("Ext2-FS [%c]: e_fscntl (cmd = %i)", dir->dev+'A', cmd));
	
	switch (cmd)
	{
		case MX_KER_XFSNAME:
		{
			strcpy ((char *) arg, "ext2-xfs");
			return E_OK;
		}
		case FS_INFO:
		{
			struct fs_info *info;
			
			info = (struct fs_info *) arg;
			if (info)
			{
				strcpy (info->name, "ext2-xfs");
				info->version = ((long) VER_MAJOR << 16) | (long) VER_MINOR;
				info->type = FS_EXT2;
				strcpy (info->type_asc, "ext2");
			}
			
			return E_OK;
		}
		case FS_USAGE:
		{
			struct fs_usage *usage;
			
			usage = (struct fs_usage *) arg;
			if (usage)
			{
				usage->blocksize   = s->sbi.s_blocksize;
				usage->blocks      = s->sbi.s_blocks_count;
				usage->free_blocks = SWAP68_L (s->sbi.s_sb->s_free_blocks_count);
				usage->inodes      = s->sbi.s_inodes_count;
				usage->free_inodes = SWAP68_L (s->sbi.s_sb->s_free_inodes_count);
			}
			
			return E_OK;
		}
		case V_CNTR_WP:
		{
			long r;
			
			r = bio.config (dir->dev, BIO_WP, arg);
			if (!r)
			{
				if (BIO_WP_CHECK (s->di) && !(s->s_flags & MS_RDONLY))
				{
					bio.sync_drv (s->di);
					
					s->s_flags |= MS_RDONLY;
					ALERT (("Ext2-FS [%c]: remounted read-only!", dir->dev+'A'));
				}
				else if (s->s_flags & MS_RDONLY)
				{
					bio.sync_drv (s->di);
					
					s->s_flags &= ~MS_RDONLY;
					ALERT (("Ext2-FS [%c]: remounted read/write!", dir->dev+'A'));
				}
			}
			
			return r;
		}
		case V_CNTR_WB:
		{
			return bio.config (dir->dev, BIO_WB, arg);
		}
		case FUTIME:
		{
			COOKIE *c;
			ushort *timeptr = (ushort *) arg;
			int uid = Geteuid ();
			fcookie fc;
			
			{
				long r;
				
				r = e_lookup (dir, name, &fc);
				if (r) return r;
				
				c = (COOKIE *) fc.index;
			}
			
			DEBUG (("Ext2-FS [%c]: e_fscntl (FUTIME) on #%li", c->dev+'A', c->inode));
			
			/* only the owner or super-user can touch
			 */
			if (uid && uid != SWAP68_W (c->in.i_uid))
			{
				e_release (&fc);
				return EACCDN;
			}
			
			if ((s->s_flags & MS_RDONLY)
				|| IS_IMMUTABLE (c))
			{
				e_release (&fc);
				return EWRPRO;
			}
			
			c->in.i_ctime = SWAP68_L (CURRENT_TIME);
			
			if (timeptr)
			{
				c->in.i_atime = SWAP68_L (unixtime (timeptr[0], timeptr[1]));
				c->in.i_mtime = SWAP68_L (unixtime (timeptr[2], timeptr[3]));
			}
			else
			{
				c->in.i_mtime = c->in.i_atime = c->in.i_ctime;
			}
			
			mark_inode_dirty (c);
			e_release (&fc);
			
			bio_SYNC_DRV ((&bio), s->di);
			return E_OK;
		}
		case FTRUNCATE:
		{
			COOKIE *c;
			fcookie fc;
			
			{
				long r;
				
				r = e_lookup (dir, name, &fc);
				if (r) return r;
				
				c = (COOKIE *) fc.index;
			}
			
			DEBUG (("Ext2-FS [%c]: e_fscntl (FTRUNCATE) on #%li", c->dev+'A', c->inode));
			
			if ((s->s_flags & MS_RDONLY)
				|| IS_IMMUTABLE (c))
			{
				e_release (&fc);
				return EWRPRO;
			}
			
			if (SWAP68_L (c->in.i_size) < arg)
			{
				e_release (&fc);
				return EACCDN;
			}
			
			c->in.i_size = SWAP68_L (arg);
			mark_inode_dirty (c);
			
			ext2_truncate (c);
			e_release (&fc);
			
			bio_SYNC_DRV ((&bio), s->di);
			return E_OK;
		}
	}
	
	DEBUG (("Ext2-FS: e_fscntl: invalid cmd or not supported"));
	return EINVFN;
}

static long _cdecl
e_dskchng (int drv, int mode)
{
	SI *s = super [drv];
	long change = 1;
	
	DEBUG (("Ext2-FS [%c]: e_dskchng (mode = %i)", drv+'A', mode));
	
	if (mode == 0)
	{
		change = BIO_DSKCHNG (s->di);
	}
	
	if (change == 0)
	{
		/* no change */
		DEBUG (("e_dskchng: leave no change"));
		return change;
	}
	
	DEBUG (("Ext2-FS [%c]: e_dskchng: invalidate drv (change = %li, memory = %li)", drv+'A', change, memory));
	
# if 0
	/* sync the inode cache */
	sync_cookies ();
# endif
	
	/* additional sync */
	bio.sync_drv (s->di);
	
	/* free the DI (invalidate also the cache units) */
	bio.free_di (s->di);
	
	/* clear directory cache */
	d_inv_dir (drv);
	/* clear inode cache */
	inv_ctable (drv);
	
	kfree (s->sbi.s_group_desc, s->sbi.s_group_desc_size);
	kfree (s, sizeof (*s));
	
	super [drv] = NULL;
	
	DEBUG (("e_dskchng: leave (change = %li, memory = %li)", change, memory));
	return change;
}

static long _cdecl
e_release (fcookie *fc)
{
	register COOKIE *c = (COOKIE *) fc->index;
# ifdef EXT2FS_DEBUG
	register long links = c->links;
# endif
	
	DEBUG (("Ext2-FS [%c]: e_release: #%li : %li", fc->dev+'A', c->inode, links, c->links));
	
	rel_cookie (c);
	
	DEBUG (("Ext2-FS [%c]: e_release: ok", fc->dev+'A'));
	return E_OK;
}

static long _cdecl
e_dupcookie (fcookie *dst, fcookie *src)
{
	((COOKIE *) src->index)->links++;
	*dst = *src;
	
	return E_OK;
}

static long _cdecl
e_sync (void)
{
	/* sync the inode cache */
	sync_cookies ();
	
	/* buffer cache automatically synced */
	return E_OK;
}
