/*
 * Filename:     inode.c
 * Project:      ext2 file system driver for MiNT
 * 
 * Note:         Please send suggestions, patches or bug reports to me
 *               or the MiNT mailing list (mint@fishpool.com).
 * 
 * Copying:      Copyright 1999 Frank Naumann (fnaumann@cs.uni-magdeburg.de)
 *               Copyright 1998, 1999 Axel Kaiser (DKaiser@AM-Gruppe.de)
 * 
 * Portions copyright 1992, 1993, 1994, 1995 Remy Card (card@masi.ibp.fr)
 * and 1991, 1992 Linus Torvalds (torvalds@klaava.helsinki.fi)
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

# include "inode.h"

# include "balloc.h"
# include "ialloc.h"
# include "super.h"
# include "truncate.h"


# define COOKIE_HASHBITS	(8)
# define COOKIE_SIZE		(1UL << COOKIE_HASHBITS)
# define COOKIE_HASHMASK	(COOKIE_SIZE - 1)

static COOKIE cookies [COOKIE_SIZE];
static COOKIE *ctable [COOKIE_SIZE];

FASTFN ulong
cookie_hash (register const ulong n)
{
	register ulong hash;
	
	hash  = n;
	hash ^= (hash >> COOKIE_HASHBITS) ^ (hash >> (COOKIE_HASHBITS << 1));
	
	return hash & COOKIE_HASHMASK;
}

FASTFN COOKIE *
cookie_lookup (register const ulong inode, register const ushort dev)
{
	register const ulong hashval = cookie_hash (inode);
	register COOKIE *c;
	
	DEBUG (("cookie_lookup: %li, inode = %li", hashval, inode));
	
	for (c = ctable [hashval]; c != NULL; c = c->next)
	{
		if ((c->dev == dev) && (c->inode == inode))
		{
			DEBUG (("cookie_lookup: match!"));
			return c;
		}
	}
	
	return NULL;
}

void
cookie_install (register COOKIE *c)
{
	register const ulong hashval = cookie_hash (c->inode);
	
	DEBUG (("cookie_install: %li, inode = %li", hashval, c->inode));
	
	c->next = ctable [hashval];
	ctable [hashval] = c;
}

FASTFN void
cookie_remove (register COOKIE *c)
{
	register const ulong hashval = cookie_hash (c->inode);
	register COOKIE **temp = & ctable [hashval];
	register long flag = 1;
	
	while (*temp)
	{
		if (*temp == c)
		{
			*temp = c->next;
			flag = 0;
			break;
		}
		temp = &(*temp)->next;
	}
	
	if (flag)
	{
		ALERT (("Ext2-FS: remove from hashtable fail in: cookie_remove (addr = %lx, %li)", c, c->inode));
	}
}

COOKIE *
get_new_cookie (void)
{
	register long i;
	
	for (i = 0; i < COOKIE_SIZE; i++)
	{
		static long count = 0;
		register COOKIE *c;
		
		count++;
		if (count == COOKIE_SIZE) count = 0;
		
		c = &(cookies [count]);
		if (!c->links)
		{
			if (c->inode)
			{
				register long r;
			
				r = SYNC_COOKIE (c);
				if (r)
				{
					ALERT (("Ext2-FS: failed to update inode #%li (r = %li)", c->inode, r));
					continue;
				}
				
				del_cookie (c);
			}
			
			memset (c, 0, sizeof (*c));
			
			c->links = 1;
			return c;
		}
	}
	
	ALERT (("Ext2-FS: get_new_cookie: no free COOKIE found"));
	
	return NULL;
}

COOKIE *
new_cookie (const ulong inode)
{
	COOKIE *c;
	
	c = get_new_cookie ();
	if (c)
	{
		c->inode = inode;
		
		cookie_install (c);
	}
	
	return c;
}

void
del_cookie (COOKIE *c)
{
	if (c->inode)
	{
		if (c->dirty)
			ALERT (("Ext2-FS [%c]: del_cookie: Inode #%li not written back!", c->dev+'A', c->inode));
		
		cookie_remove (c);
	}
	
	c->links = 0;
	c->inode = 0;
	c->dirty = 0;
	
	if (c->open)
	{
		ALERT (("Ext2-FS: open FILEPTR detect in: del_cookie #%li", c->inode));
		c->open = NULL;
	}
	if (c->locks)
	{
		ALERT (("Ext2-FS: open LOCKS detect in: del_cookie #%li", c->inode));
		c->locks = NULL;
	}
	
	clear_lastlookup (c);
}

void
inv_ctable (ushort drv)
{
	register long i;
	
	for (i = 0; i < COOKIE_SIZE; i++)
	{
		register COOKIE *c;
		
		for (c = ctable [i]; c != NULL; c = c->next)
		{
			if (c->dev == drv && c->inode)
			{
				del_cookie (c);
			}
		}
	}
}

long
get_cookie (SI *s, long inode, COOKIE **in)
{
	COOKIE *c;
	
	DEBUG (("Ext2-FS [%c]: get_cookie enter: #%li", s->dev+'A', inode));
	
	c = cookie_lookup (inode, s->dev);
	if (c)
	{
		*in = c;
		c->links++;
		
		DEBUG (("Ext2-FS: get_cookie leave ok, found in cache"));
		return E_OK;
	}
	
	c = new_cookie (inode);
	if (c)
	{
		ext2_gd *i_g;
		long group;
		long offset;
		long block;
		UNIT *u;
		
		if ((inode != EXT2_ROOT_INO
				&& inode != EXT2_ACL_IDX_INO
				&& inode != EXT2_ACL_DATA_INO
				&& inode < EXT2_FIRST_INO (s))
			|| inode > (s->sbi.s_inodes_count))
		{
			ALERT (("ext2_read_inode: bad inode number: #%li", inode));
			
			del_cookie (c);
			
			DEBUG (("Ext2-FS: get_cookie: leave failure (inode out of range [%li])", inode));
			return ERANGE;
		}
		
		/* inodes are numbered from 1
		 * -> C offset correction
		 */
		inode--;
		
		/* group number */
		group = inode / EXT2_INODES_PER_GROUP (s);
		
		/* index on group */
		offset = inode % EXT2_INODES_PER_GROUP (s) * EXT2_INODE_SIZE (s);
		
		/* ptr to the actual group descriptor */
		i_g = ext2_get_group_desc (s, group, NULL);
		
		/* block of the inode */
		block  = SWAP68_L (i_g->bg_inode_table);
		block += (offset >> EXT2_BLOCK_SIZE_BITS (s));
		
		u = bio.read (s->di, block, s->sbi.s_blocksize);
		if (u)
		{
			offset &= EXT2_BLOCK_SIZE_MASK (s);
			
			
			/* setup cookie
			 */
			
			c->dirty = 0;
			
			c->s = s;
			c->i_flags = s->s_flags;
			
			c->dev = s->dev;
			c->rdev = s->rdev;
			
			c->open = NULL;
			c->locks = NULL;
			c->lastlookup = NULL;
			c->lastlookupsize = 0;
			
			/* copy inode */
			c->in = *((ext2_in *) (u->data + offset));
			
			c->i_block_group = group;
			c->i_next_alloc_block = 0;
			c->i_next_alloc_goal = 0;
			c->i_prealloc_block = 0;
			c->i_prealloc_count = 0;
			
			{
				register ulong i_flags;
				
				i_flags = SWAP68_W (c->in.i_flags);
				
				if (i_flags & EXT2_SYNC_FL)
					c->i_flags |= MS_SYNCHRONOUS;
				
				if (i_flags & EXT2_APPEND_FL)
					c->i_flags |= S_APPEND;
				
				if (i_flags & EXT2_IMMUTABLE_FL)
					c->i_flags |= S_IMMUTABLE;
				
				if (i_flags & EXT2_NOATIME_FL)
					c->i_flags |= MS_NOATIME;
			}
			
			*in = c;
			
			DEBUG (("Ext2-FS: get_cookie leave ok"));
			return E_OK;
		}
		
		del_cookie (c);
	}
	
	DEBUG (("Ext2-FS: get_cookie: leave failure (new_cookie fail)"));
	return ENSMEM;
}

long
put_cookie (COOKIE *c)
{
	SI *s = c->s;
	long inode = c->inode;
	
	ext2_gd *i_g;
	long group;
	long offset;
	long block;
	UNIT *u;
	
	
	DEBUG (("Ext2-FS [%c]: put_inode enter (%li)", c->dev+'A', inode));
	
	if ((inode != EXT2_ROOT_INO
			&& inode != EXT2_ACL_IDX_INO
			&& inode != EXT2_ACL_DATA_INO
			&& inode < EXT2_FIRST_INO (s))
		|| inode > (s->sbi.s_inodes_count))
	{
		ALERT (("ext2_write_inode: bad inode number: #%li", inode));
		
		/* bad inode */
		
		DEBUG (("Ext2-FS: read_inode: leave failure (inode out of range [%li])", inode));
		return ERANGE;
	}
	
	/* inodes are numbered from 1
	 * -> C offset correction
	 */
	inode--;
	
	/* group number */
	group = inode / EXT2_INODES_PER_GROUP (s);
	
	/* index on group */
	offset = inode % EXT2_INODES_PER_GROUP (s) * EXT2_INODE_SIZE (s);
	
	/* ptr to the actual group descriptor */
	i_g = ext2_get_group_desc (s, group, NULL);
	
	/* block of the inode */
	block  = SWAP68_L (i_g->bg_inode_table);
	block += (offset >> EXT2_BLOCK_SIZE_BITS (s));
	
	/* read target block */
	u = bio.read (s->di, block, s->sbi.s_blocksize);
	if (u)
	{
		offset &= EXT2_BLOCK_SIZE_MASK (s);
		
		/* copy the inode */
		*((ext2_in *) (u->data + offset)) = c->in;
		
		/* and write back */
		bio_MARK_MODIFIED (&bio, u);
		
		DEBUG (("Ext2-FS: put_inode leave ok"));
		return E_OK;
	}
	
	DEBUG (("Ext2-FS: out_inode: leave failure (bio.read fail)"));
	return EREADF;
}

void
rel_cookie (COOKIE *c)
{
	c->links--;
	
	if (!c->links)
	{
		if (!c->in.i_links_count && !c->open)
		{
			DEBUG (("Ext2-FS [%c]: rel_cookie: free deleted inode #%li", c->dev+'A', c->inode));
			ext2_delete_inode (c);
		}
	}
}

void
sync_cookies (void)
{
	long i;
	
	for (i = 0; i < COOKIE_SIZE; i++)
	{
		register COOKIE *c = &(cookies [i]);
		
		if (c->inode)
		{
			register long r;
			
			r = SYNC_COOKIE (c);
			if (r)
				ALERT (("Ext2-FS: failed to update inode #%li (r = %li)", c->inode, r));
		}
	}
}

void
ext2_delete_inode (COOKIE *inode)
{
	if (inode->inode == EXT2_ACL_IDX_INO
		|| inode->inode == EXT2_ACL_DATA_INO)
	{
		return;
	}
	
	inode->in.i_dtime = SWAP68_L (CURRENT_TIME);
	inode->in.i_size = 0;
	mark_inode_dirty (inode);
	
	if (inode->in.i_blocks)
		ext2_truncate (inode);
	
	ext2_free_inode (inode);
}


FASTFN ulong
inode_bmap (COOKIE *inode, ulong nr)
{
	return SWAP68_L (inode->in.i_block [nr]);
}

FASTFN ulong
block_bmap (UNIT *u, ulong nr)
{
	if (!u)
		return 0;
	
	return SWAP68_L (((ulong *) u->data) [nr]);
}

ulong
ext2_bmap (COOKIE *inode, long block)
{
	long addr_per_block = EXT2_ADDR_PER_BLOCK (inode->s);
	long addr_per_block_bits = EXT2_ADDR_PER_BLOCK_BITS (inode->s);
	long i;
	
	DEBUG (("Ext2-FS: ext2_bmap enter (%li)", block));
	
	if (block < 0)
	{
		ALERT (("ext2_bmap: block < 0"));
		return 0;
	}
	
	if (block >= EXT2_NDIR_BLOCKS + addr_per_block +
		(1UL << (addr_per_block_bits * 2)) +
		((1UL << (addr_per_block_bits * 2)) << addr_per_block_bits))
	{
		ALERT (("ext2_bmap: block (%li) > big (%li)",
			block,
			EXT2_NDIR_BLOCKS + addr_per_block +
			(1UL << (addr_per_block_bits * 2)) +
			((1UL << (addr_per_block_bits * 2)) << addr_per_block_bits)
		));
		return 0;
	}
	
	if (block < EXT2_NDIR_BLOCKS)
		return inode_bmap (inode, block);
	
	block -= EXT2_NDIR_BLOCKS;
	if (block < addr_per_block)
	{
		i = inode_bmap (inode, EXT2_IND_BLOCK);
		if (!i)
			return 0;
		
		return block_bmap (bio.read (inode->s->di, i, inode->s->sbi.s_blocksize), block);
	}
	
	block -= addr_per_block;
	if (block < (1UL << (addr_per_block_bits * 2)))
	{
		i = inode_bmap (inode, EXT2_DIND_BLOCK);
		if (!i)
			return 0;
		
		i = block_bmap (bio.read (inode->s->di, i, inode->s->sbi.s_blocksize), block >> addr_per_block_bits);
		if (!i)
			return 0;
		
		return block_bmap (bio.read (inode->s->di, i, inode->s->sbi.s_blocksize), block & (addr_per_block - 1));
	}
	
	block -= (1UL << (addr_per_block_bits * 2));
	i = inode_bmap (inode, EXT2_TIND_BLOCK);
	if (!i)
		return 0;
	
	i = block_bmap (bio.read (inode->s->di, i, inode->s->sbi.s_blocksize), block >> (addr_per_block_bits * 2));
	if (!i)
		return 0;
	
	i = block_bmap (bio.read (inode->s->di, i, inode->s->sbi.s_blocksize), (block >> addr_per_block_bits) & (addr_per_block - 1));
	if (!i)
		return 0;
	
	return block_bmap (bio.read (inode->s->di, i, inode->s->sbi.s_blocksize), block & (addr_per_block - 1));
}

UNIT *
ext2_read (COOKIE *inode, long block, long *err)
{
	ulong tmp;
	
	tmp = ext2_bmap (inode, block);
	if (tmp)
		return bio.read (inode->s->di, tmp, inode->s->sbi.s_blocksize);
	
	if (err)
		*err = EREADF;
	
	return NULL;
}


/*  ext2_discard_prealloc and ext2_alloc_block are atomic wrt. the
 * superblock in the same manner as are ext2_free_blocks and
 * ext2_new_block.  We just wait on the super rather than locking it
 * here, since ext2_new_block will do the necessary locking and we
 * can't block until then.
 */
void
ext2_discard_prealloc (COOKIE *inode)
{
# ifdef EXT2_PREALLOCATE
	register ushort total;
	
	DEBUG (("ext2_discard_prealloc: enter (#%li)", inode->inode));
	
	if (inode->i_prealloc_count)
	{
		total = inode->i_prealloc_count;
		inode->i_prealloc_count = 0;
		ext2_free_blocks (inode, inode->i_prealloc_block, total);
	}
	
	DEBUG (("ext2_discard_prealloc: leave"));
# endif
}

static long
ext2_alloc_block (COOKIE *inode, ulong goal, long *err)
{
# ifdef EXT2FS_DEBUG
	static ulong alloc_hits = 0;
	static ulong alloc_attempts = 0;
# endif
	ulong result;
	
	
	DEBUG (("ext2_alloc_block: enter"));
	DEBUG (("ext2_alloc_block: inode = %li, goal = %ld", inode->inode, goal));
	
	/* wait_on_super (inode->i_sb); */
	
# ifdef EXT2_PREALLOCATE
	if (inode->i_prealloc_count
		&& (goal == inode->i_prealloc_block
			|| goal + 1 == inode->i_prealloc_block))
	{		
		result = inode->i_prealloc_block++;
		inode->i_prealloc_count--;
		
		DEBUG (("preallocation hit (%lu/%lu)", ++alloc_hits, ++alloc_attempts));
	}
	else
	{
		ext2_discard_prealloc (inode);
		
		DEBUG (("preallocation miss (%lu/%lu)", alloc_hits, ++alloc_attempts));
		
		if (EXT2_ISREG (SWAP68_W (inode->in.i_mode)))
		{
			result = ext2_new_block (inode, goal, 
					&inode->i_prealloc_count,
					&inode->i_prealloc_block, err);
		}
		else
		{
			result = ext2_new_block (inode, goal, 0, 0, err);
		}
	}
# else
	result = ext2_new_block (inode, goal, 0, 0, err);
# endif
	
	DEBUG (("ext2_alloc_block: leave (result = %ld, *err = %li)", result, *err));
	return result;
}

static ulong
inode_getblk (COOKIE *inode, long nr, long new_block, long *err, ushort clear_flag)
{
	ulong *p = inode->in.i_block + nr;
	ulong tmp = *p;
	ulong goal;
	
	DEBUG (("inode_getblk: enter (#%li, nr = %li, new_block = %li)", inode->inode, nr, new_block));
	
	if (tmp)
		return SWAP68_L (tmp);
	
	*err = EFBIG;
	/* Check file limits */
	{
	}
	
	if (inode->i_next_alloc_block == new_block)
		goal = inode->i_next_alloc_goal;
	else
		goal = 0;
	
	DEBUG (("hint = %ld,", goal));
	
	if (!goal)
	{
		for (tmp = nr - 1; tmp >= 0; tmp--)
		{
			if (inode->in.i_block [tmp])
			{
				goal = SWAP68_L (inode->in.i_block [tmp]);
				break;
			}
		}
		
		if (!goal)
			goal = (inode->i_block_group * EXT2_BLOCKS_PER_GROUP (inode->s)) +
				inode->s->sbi.s_first_data_block;
	}
	
	DEBUG (("goal = %ld", goal));
	
	tmp = ext2_alloc_block (inode, goal, err);
	if (tmp)
	{
		ulong blocks = inode->s->sbi.s_blocksize >> 9;
		
		*p = SWAP68_L (tmp);
		inode->i_next_alloc_block = new_block;
		inode->i_next_alloc_goal = tmp;
		inode->in.i_ctime = SWAP68_L (CURRENT_TIME);
		inode->in.i_blocks = SWAP68_L (SWAP68_L (inode->in.i_blocks) + blocks);
		mark_inode_dirty (inode);
		
		if (clear_flag)
		{
			UNIT *u;
			
			u = bio.getunit (inode->s->di, tmp, inode->s->sbi.s_blocksize);
			if (u)
				memset (u->data, 0, inode->s->sbi.s_blocksize);
			else
			{
				tmp = 0;
			}
		}
	}
	
	DEBUG (("inode_getblk: leave (tmp = %ld)", tmp));
	return tmp;
}

static long
block_getblk (COOKIE *inode, long block, long nr, long blocksize, long new_block, long *err, ushort clear_flag)
{
	ulong tmp;
	ulong goal;
	UNIT *u;
	
	DEBUG (("block_getblk: enter (#%li, block = %li, nr = %li)", inode->inode, block, nr));
	DEBUG (("block_getblk: blocksize = %li, new_block = %li", blocksize, new_block));
	
	if (!block)
	{
		DEBUG (("block_getblk: leave (!block -> 0)"));
		return 0;
	}
	
	u = bio.read (inode->s->di, block, blocksize);
	if (!u)
	{
		DEBUG (("block_getblk: leave (!u1 (%li, %li) -> 0)", block, blocksize));
		return 0;
	}
	
	tmp = *((ulong *) u->data + nr);
	if (tmp)
	{
		return SWAP68_L (tmp);
	}
	
	*err = EFBIG;
	/* Check file limits */
	{
	}
	
	if (inode->i_next_alloc_block == new_block)
		goal = inode->i_next_alloc_goal;
	else
		goal = 0;
	
	if (!goal)
	{
		for (tmp = nr - 1; tmp >= 0; tmp--)
		{
			if (((ulong *) u->data) [tmp])
			{
				goal = SWAP68_L (((ulong *) u->data) [tmp]);
				break;
			}
		}
		
		if (!goal)
			goal = u->sector;
	}
	
	tmp = ext2_alloc_block (inode, goal, err);
	if (!tmp)
	{
		DEBUG (("block_getblk: leave (!tmp -> 0)"));
		return 0;
	}
	
	u = bio.read (inode->s->di, block, blocksize);
	if (!u)
	{
		DEBUG (("block_getblk: leave (!u2 (%li, %li) -> 0)", block, blocksize));
		return 0;
	}
	
	*((ulong *) u->data + nr) = SWAP68_L (tmp);
	bio_MARK_MODIFIED (&bio, u);
	
	inode->i_next_alloc_block = new_block;
	inode->i_next_alloc_goal = tmp;
	inode->in.i_ctime = SWAP68_L (CURRENT_TIME);
	inode->in.i_blocks = SWAP68_L (SWAP68_L (inode->in.i_blocks) + (blocksize >> 9));
	mark_inode_dirty (inode);
	
	if (clear_flag)
	{
		u = bio.getunit (inode->s->di, tmp, inode->s->sbi.s_blocksize);
		if (u)
			memset (u->data, 0, inode->s->sbi.s_blocksize);
		else
		{
			tmp = 0;
		}
	}
	
	DEBUG (("block_getblk: leave (tmp = %li)", tmp));
	return tmp;
}

long
ext2_getblk (COOKIE *inode, long block, long *err, ushort clear_flag)
{
	ulong addr_per_block = EXT2_ADDR_PER_BLOCK (inode->s);
	ulong addr_per_block_bits = EXT2_ADDR_PER_BLOCK_BITS (inode->s);
	ulong b = block;
	ulong u;	
	
	*err = EIO;
	
	DEBUG (("ext2_getblk: enter (#%li, block = %li)", inode->inode, block));
	
	if (block < 0)
	{
		ALERT (("ext2_getblk: block < 0"));
		return 0;
	}
	
	if (block > EXT2_NDIR_BLOCKS + addr_per_block +
		(1UL << (addr_per_block_bits * 2)) +
		((1UL << (addr_per_block_bits * 2)) << addr_per_block_bits))
	{
		ALERT (("ext2_getblk: block > big"));
		return 0;
	}
	
	/* If this is a sequential block allocation, set the next_alloc_block
	 * to this block now so that all the indblock and data block
	 * allocations use the same goal zone
	 */
	
	DEBUG (("block %lu, next %lu, goal %lu", block, 
		inode->i_next_alloc_block,
		inode->i_next_alloc_goal));
	
	if (block == inode->i_next_alloc_block + 1)
	{
		inode->i_next_alloc_block++;
		inode->i_next_alloc_goal++;
	}
	
	*err = ENOSPC;
	
	if (block < EXT2_NDIR_BLOCKS)
	{
		DEBUG (("ext2_getblk: leave (return inode_getblk)"));
		return inode_getblk (inode, block, b, err, clear_flag);
	}
	
	block -= EXT2_NDIR_BLOCKS;
	if (block < addr_per_block)
	{
		u = inode_getblk (inode, EXT2_IND_BLOCK, b, err, 1);
		
		DEBUG (("ext2_getblk: leave (return block_getblk 1)"));
		return block_getblk (inode, u, block, inode->s->sbi.s_blocksize, b, err, clear_flag);
	}
	
	block -= addr_per_block;
	if (block < (1UL << (addr_per_block_bits * 2)))
	{
		u = inode_getblk (inode, EXT2_DIND_BLOCK, b, err, 1);
		u = block_getblk (inode, u, block >> addr_per_block_bits, inode->s->sbi.s_blocksize, b, err, 1);
		
		DEBUG (("ext2_getblk: leave (return block_getblk 2)"));
		return block_getblk (inode, u, block & (addr_per_block - 1), inode->s->sbi.s_blocksize, b, err, clear_flag);
	}
	
	block -= (1UL << (addr_per_block_bits * 2));
	u = inode_getblk (inode, EXT2_TIND_BLOCK, b, err, 1);
	u = block_getblk (inode, u, block >> (addr_per_block_bits * 2), inode->s->sbi.s_blocksize, b, err, 1);
	u = block_getblk (inode, u, (block >> addr_per_block_bits) & (addr_per_block - 1), inode->s->sbi.s_blocksize, b, err, 1);
	
	DEBUG (("ext2_getblk: leave (return block_getblk 3)"));
	return block_getblk (inode, u, block & (addr_per_block - 1), inode->s->sbi.s_blocksize, b, err, clear_flag);
}

UNIT *
ext2_bread (COOKIE *inode, long block, long *err)
{
	ulong prev_blocks = SWAP68_L (inode->in.i_blocks);
	ulong nr;
	UNIT *u;
	
	DEBUG (("ext2_bread: enter (#%li, block = %li)", inode->inode, block));
	
	nr = ext2_getblk (inode, block, err, 1);
	if (!nr)
	{
		DEBUG (("ext2_bread: leave (!nr -> NULL)"));
		return NULL;
	}
	
	/* If the inode has grown, and this is a directory, then perform
	 * preallocation of a few more blocks to try to keep directory
	 * fragmentation down.
	 */
	if (EXT2_ISDIR (SWAP68_W (inode->in.i_mode))
		&& SWAP68_L (inode->in.i_blocks) > prev_blocks
		&& EXT2_HAS_COMPAT_FEATURE (inode->s, EXT2_FEATURE_COMPAT_DIR_PREALLOC))
	{
		ulong tmp;
		long i;
		
		for (i = 1; i < EXT2_SB (inode->s)->s_prealloc_dir_blocks; i++)
		{
			tmp = ext2_getblk (inode, block + i, err, 1);
			if (!tmp)
			{
				ext2_free_blocks (inode, block + i, 1);
				break;
			}
		}
	}
	
	u = bio.read (inode->s->di, nr, inode->s->sbi.s_blocksize);
	
	DEBUG (("ext2_bread: leave (%li, %li -> %lx)", nr, inode->s->sbi.s_blocksize, u));
	return u;
}
