/*
 * Filename:     namei.c
 * Project:      ext2 file system driver for MiNT
 * 
 * Note:         Please send suggestions, patches or bug reports to me
 *               or the MiNT mailing list (mint@fishpool.com).
 * 
 * Copying:      Copyright 1999 Frank Naumann (fnaumann@cs.uni-magdeburg.de)
 *               Copyright 1998, 1999 Axel Kaiser (DKaiser@AM-Gruppe.de)
 * 
 * Portions copyright 1992, 1993, 1994, 1995 Remy Card (card@masi.ibp.fr)
 * and 1991, 1992 Linus Torvalds (torvalds@klaava.helsinki.fi)
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

# include "namei.h"

# include "inode.h"
# include "super.h"


# define CACHE_DIR_HASHBITS	(9)
# define CACHE_DIR_SIZE		(1UL << CACHE_DIR_HASHBITS)
# define CACHE_DIR_HASHMASK	(CACHE_DIR_SIZE - 1)

static _DIR chache_dir [CACHE_DIR_SIZE];
static _DIR * dtable [CACHE_DIR_SIZE];

/* ?
 *	prevhash = (prevhash << 4) | (prevhash >> 28);
 *	return prevhash ^ _TOUPPER (c);
 */

FASTFN ulong
d_hash_hash (register const char *s)
{
	register ulong hash = 0;
	
	while (*s)
	{
		hash = ((hash << 5) - hash) + _TOUPPER (*s);
		s++;
	}
	
	hash ^= (hash >> CACHE_DIR_HASHBITS) ^ (hash >> (CACHE_DIR_HASHBITS << 1));
	
	return hash & CACHE_DIR_HASHMASK;
}

FASTFN void
d_hash_setup (_DIR *dentry, const char *name, long namelen)
{
	register ulong hash = 0;
	register char *s;
	
	dentry->namelen = namelen;
	dentry->name = s = kmalloc (namelen + 1);
	
	while (namelen--)
	{
		*s = *name++;
		hash = ((hash << 5) - hash) + _TOUPPER (*s);
		s++;
	}
	
	*s = '\0';
	
	hash ^= (hash >> CACHE_DIR_HASHBITS) ^ (hash >> (CACHE_DIR_HASHBITS << 1));
	
	dentry->hash = (hash & CACHE_DIR_HASHMASK);
	
	DEBUG (("Ext2-FS: %s, %i, %i", dentry->name, dentry->namelen, dentry->hash));
}

FASTFN _DIR *
d_hash_lookup (register const char *s, register ulong parent, register ushort dev)
{
	register _DIR *dentry;
	
	DEBUG (("d_hash_lookup: %li, %s", d_hash_hash (s), s));
	
	for (dentry = dtable [d_hash_hash (s)]; dentry != NULL; dentry = dentry->next)
	{
		if ((dentry->dev == dev)
			&& (dentry->parent == parent)
			&& (_STRCMP (dentry->name, s) == 0))
		{
			DEBUG (("d_hash_lookup: match s = %s, dentry->name = %s", s, dentry->name));
			return dentry;
		}
	}
	
	return NULL;
}

FASTFN void
d_hash_install (register _DIR *dentry)
{
	register const ulong hashval = dentry->hash;
	
	DEBUG (("c_hash_install: %i, %s", dentry->hash, dentry->name));
	
	dentry->next = dtable [hashval];
	dtable [hashval] = dentry;
}

FASTFN void
d_hash_remove  (register _DIR *dentry)
{
	register const ulong hashval = dentry->hash;
	register _DIR **temp = & dtable [hashval];
	register long flag = 1;
	
	while (*temp)
	{
		if (*temp == dentry)
		{
			*temp = dentry->next;
			flag = 0;
			break;
		}
		temp = &(*temp)->next;
	}
	
	if (flag)
	{
		ALERT (("Ext2-FS: remove from hashtable fail in: d_hash_remove (addr = %lx, %s)", dentry, dentry->name));
	}
}

_DIR *
d_lookup (COOKIE *dir, const char *name)
{
	return d_hash_lookup (name, dir->inode, dir->dev);
}

_DIR *
d_get_dir (COOKIE *dir, ulong inode, const char *name, long namelen)
{
	register long i;
	
	for (i = 0; i < CACHE_DIR_SIZE; i++)
	{
		static long count = 0;
		register _DIR *dentry;
		
		count++;
		if (count == CACHE_DIR_SIZE)
			count = 0;
		
		dentry = &(chache_dir [count]);
		if (dentry->links == 0)
		{
			if (dentry->name)
				d_del_dir (dentry);
			
			d_hash_setup (dentry, name, namelen);
			d_hash_install (dentry);
			
			dentry->links = 1;
			dentry->parent = dir->inode;
			dentry->inode = inode;
			dentry->dev = dir->dev;
			
			return dentry;
		}
	}
	
	ALERT (("Ext2-FS: d_get_dir: no free _DIR found for %s", name));
	return NULL;
}

void
d_del_dir (_DIR *dentry)
{
	d_hash_remove (dentry);
	
	kfree (dentry->name, dentry->namelen + 1);
	dentry->name = NULL;
}

void
d_inv_dir (ushort dev)
{
	long i;
	
	for (i = 0; i < CACHE_DIR_SIZE; i++)
	{
		_DIR *dentry;
		
		dentry = &(chache_dir [i]);
		if (dentry->dev == dev)
		{
			if (dentry->name)
				d_del_dir (dentry);
		}
	}
}


long
ext2_check_dir_entry (const char *caller, COOKIE *dir, ext2_d2 *de, UNIT *u, ulong offset)
{
	const char *error_msg = NULL;
	register ushort rec_len = SWAP68_W (de->rec_len);
	
	if (rec_len < EXT2_DIR_REC_LEN (1))
		error_msg = "rec_len is smaller than minimal";
	else if (rec_len % 4 != 0)
		error_msg = "rec_len % 4 != 0";
	else if (rec_len < EXT2_DIR_REC_LEN (de->name_len))
		error_msg = "rec_len is too small for name_len";
	else if (dir && ((char *) de - u->data) + rec_len > dir->s->sbi.s_blocksize)
		error_msg = "directory entry across blocks";
	else if (dir && SWAP68_L (de->inode) > SWAP68_L (dir->s->sbi.s_sb->s_inodes_count))
		error_msg = "inode out of bounds";
	
	if (error_msg)
	{
		ALERT (("%s: bad entry in directory #%lu: %s - "
			"offset=%lu, inode=%lu, rec_len=%d, name_len=%d",
			caller, dir->inode, error_msg, offset,
			SWAP68_L (de->inode), rec_len, de->name_len));
		
		return 0;
	}
	
	return 1;
}

/*
 * returns 1 for success, 0 for failure.
 *
 * `length <= EXT2_NAME_LEN' is guaranteed by caller.
 * `de != NULL' is guaranteed by caller.
 */

FASTFN long
ext2_match (long length, const char *name, ext2_d2 *de)
{
	if (length != de->name_len)
		return 0;
	
	if (de->inode == 0)
		return 0;
	
	return !memcmp (name, de->name, length);
}


_DIR *
ext2_search_entry (COOKIE *dir, const char *name, long namelen)
{
	SI *s = super [dir->dev];
	ulong size = SWAP68_L (dir->in.i_size);
	
	long block;
	long offset;
	
	_DIR *dentry;
	
	DEBUG (("Ext2-FS [%c]: ext2_search_entry (%li, %s)", dir->dev+'A', namelen, name));
	
	if (namelen == 0)
		return NULL;
	
	/* 1. search in directory cache
	 */
	dentry = d_lookup (dir, name);
	if (dentry)
	{
		DEBUG (("Ext2-FS: ext2_search_entry: found in dcache (%li, %li, %li, %s)", dir->inode, dentry->parent, dentry->inode, dentry->name));
		return dentry;
	}
	
	/* 2. check lastlookup fail cache
	 */
	if (dir->lastlookup)
	{
		if ((namelen == dir->lastlookupsize)
			&& (_STRCMP (name, dir->lastlookup) == 0))
		{
			DEBUG (("ext2_search_entry: leave not found (lastlookup) (name = %s)", name));
			return NULL;
		}
	}
	
	/* 3. search on disk
	 */
	for (block = 0, offset = 0; offset < size; block++)
	{
		UNIT *u;
		ext2_d2 *de;
		char *upper;
		
		u = ext2_read (dir, block, NULL);
		if (!u)
		{
			DEBUG (("Ext2-FS: ext2_search_entry: ext2_read fail (%li)", block));
			goto failure;
		}
		
		de = (ext2_d2 *) u->data;
		
		upper = (char *) de + s->sbi.s_blocksize;
		while ((char *) de < upper)
		{
			/* this code is executed quadratically often
			 * do minimal checking `by hand'
			 */
			long de_len;
			
			if ((char *) de + namelen <= upper && ext2_match (namelen, name, de))
			{
				/* found a match -
				 * just to be sure, do a full check
				 */
				if (!ext2_check_dir_entry ("ext2_search_entry", dir, de, u, offset))
					goto failure;
				
				dentry = d_get_dir (dir, SWAP68_L (de->inode), de->name, de->name_len);
				if (dentry)
					dentry->links--;
				
				DEBUG (("Ext2-FS: ext2_search_entry: found (%i, %s)", dentry->namelen, dentry->name));
				return dentry;
			}
			
			de_len = SWAP68_W (de->rec_len);
			if (de_len <= 0)
				goto failure;
			
			offset += de_len;
			de = (ext2_d2 *) ((char *) de + de_len);
		}
	}
	
failure:
	/* 4. update lookup fail cache
	 */
	update_lastlookup (dir, name, namelen);
	
	
	DEBUG (("Ext2-FS [%c]: ext2_search_entry: leave not found", dir->dev+'A'));
	return NULL;
}

# if 0
_DIR *
ext2_search_entry_i (COOKIE *dir, long inode)
{
	SI *s = super [dir->dev];
	ulong size = SWAP68_L (dir->in.i_size);
	
	long block;
	long offset;
	
	DEBUG (("Ext2-FS [%c]: ext2_search_entry_i: #%li", dir->dev+'A', inode));
	
	for (block = 0, offset = 0; offset < size; block++)
	{
		UNIT *u;
		ext2_d2 *de;
		char *upper;
		
		u = ext2_read (dir, block, NULL);
		if (!u)
		{
			DEBUG (("Ext2-FS: ext2_search_entry_i: ext2_read fail (%li)", block));
			goto failure;
		}
		
		de = (ext2_d2 *) u->data;
		
		upper = (char *) de + s->sbi.s_blocksize;
		while ((char *) de < upper)
		{
			/* this code is executed quadratically often
			 * do minimal checking `by hand'
			 */
			long de_len;
			
			if (inum == SWAP68_L (de->inode))
			{
				_DIR *dentry;
				
				/* found a match -
				 * just to be sure, do a full check
				 */
				if (!ext2_check_dir_entry ("ext2_search_entry_i", dir, de, u, offset))
					goto failure;
				
				/* search in dir cache
				 */
				dentry = d_lookup ();
				
				if (!dentry)
				{
					dentry = d_get_dir (dir, SWAP68_L (de->inode), de->name, de->name_len);
					if (dentry)
						dentry->links--;
				}
				
				DEBUG (("Ext2-FS: ext2_search_entry_i: found (%i, %s)", dentry->namelen, dentry->name));
				return dentry;
			}
			
			de_len = SWAP68_W (de->rec_len);
			if (de_len <= 0)
				goto failure;
			
			offset += de_len;
			de = (ext2_d2 *) ((char *) de + de_len);
		}
	}
	
failure:
	
	DEBUG (("Ext2-FS [%c]: ext2_search_entry_i: leave not found", dir->dev+'A'));
	return NULL;
}
# endif

/*
 *	ext2_find_entry()
 *
 * finds an entry in the specified directory with the wanted name. It
 * returns the cache buffer in which the entry was found, and the entry
 * itself (as a parameter - res_dir). It does NOT read the inode of the
 * entry - you'll have to do that yourself if you want to.
 */
UNIT *
ext2_find_entry (COOKIE *dir, const char *name, long namelen, ext2_d2 **res_dir)
{
	SI *s;
	ulong size;
	
	long block;
	long offset;
	
	DEBUG (("Ext2-FS [%c]: ext2_find_entry (%li, %s)", dir->dev+'A', namelen, name));
	
	*res_dir = NULL;
	if (!dir)
		return NULL;
	
	if (namelen == 0)
		return NULL;
	
	s = super [dir->dev];
	size = SWAP68_L (dir->in.i_size);
	
	for (block = 0, offset = 0; offset < size; block++)
	{
		UNIT *u;
		ext2_d2 *de;
		char *upper;
		
		u = ext2_read (dir, block, NULL);
		if (!u)
		{
			DEBUG (("Ext2-FS: ext2_find_entry: ext2_read fail (%li)", block));
			goto failure;
		}
		
		de = (ext2_d2 *) u->data;
		
		upper = (char *) de + s->sbi.s_blocksize;
		while ((char *) de < upper)
		{
			/* this code is executed quadratically often
			 * do minimal checking `by hand'
			 */
			long de_len;
			
			if ((char *) de + namelen <= upper && ext2_match (namelen, name, de))
			{
				/* found a match -
				 * just to be sure, do a full check
				 */
				if (!ext2_check_dir_entry ("ext2_find_entry", dir, de, u, offset))
					goto failure;
				
				*res_dir = de;
				return u;
			}
			
			de_len = SWAP68_W (de->rec_len);
			if (de_len <= 0)
				goto failure;
			
			offset += de_len;
			de = (ext2_d2 *) ((char *) de + de_len);
		}
	}
	
failure:
	DEBUG (("Ext2-FS [%c]: ext2_find_entry: leave not found", dir->dev+'A'));
	return NULL;
}

/*
 *	ext2_add_entry()
 *
 * adds a file entry to the specified directory, using the same
 * semantics as ext2_find_entry(). It returns NULL if it failed.
 *
 * NOTE!! The inode part of 'de' is left at 0 - which means you
 * may not sleep between calling this and putting something into
 * the entry, as someone else might have used it while you slept.
 */
UNIT *
ext2_add_entry (COOKIE *dir, const char *name, long namelen, ext2_d2 **res_dir, long *err)
{
	SI *s;
	UNIT *u;
	ext2_d2 *de, *de1;
	
	ulong offset;
	ushort rec_len;
	
	*res_dir = NULL;
	
	if (!dir || !dir->in.i_links_count)
	{
		*err = EINVAL;
		return NULL;
	}
	
	s = dir->s;
	
	if (namelen > EXT2_NAME_LEN)
	{
		*err = ENAMETOOLONG;
		return NULL;
	}
	
	if (!namelen)
	{
		*err = EACCDN;
		return NULL;
	}
	
	/* Is this a busy deleted directory?
	 * Can't create new files if so
	 */
	if (!dir->in.i_size)
	{
		*err = ENOENT;
		return NULL;
	}
	
	u = ext2_read (dir, 0, err);
	if (!u)
	{
		*err = ENOSPC;
		return NULL;
	}
	
	rec_len = EXT2_DIR_REC_LEN (namelen);
	offset = 0;
	de = (ext2_d2 *) u->data;
	
	while (1)
	{
		if ((char *) de >= s->sbi.s_blocksize + u->data)
		{
			u = ext2_bread (dir, offset >> EXT2_BLOCK_SIZE_BITS (s), err);
			if (!u)
			{
				*err = ENOSPC;
				return NULL;
			}
			
			if (SWAP68_L (dir->in.i_size) <= offset)
			{
				if (!dir->in.i_size)
				{
					*err = ENOENT;
					return NULL;
				}
				
				DEBUG (("creating next block"));
				
				de = (ext2_d2 *) u->data;
				de->inode = 0;
				de->rec_len = SWAP68_W (s->sbi.s_blocksize);
				dir->in.i_size = SWAP68_L (offset + s->sbi.s_blocksize);
				dir->in.i_flags = SWAP68_L (SWAP68_L (dir->in.i_flags) & ~EXT2_BTREE_FL);
				mark_inode_dirty (dir);
			}
			else
			{
				DEBUG (("skipping to next block"));
				
				de = (ext2_d2 *) u->data;
			}
		}
		
		if (!ext2_check_dir_entry ("ext2_add_entry", dir, de, u, offset))
		{
			*err = ENOENT;
			return NULL;
		}
		
		if (ext2_match (namelen, name, de))
		{
			*err = EEXIST;
			return NULL;
		}
		
		if ((de->inode == 0 && SWAP68_W (de->rec_len) >= rec_len)
			|| (SWAP68_W (de->rec_len) >= EXT2_DIR_REC_LEN (de->name_len) + rec_len))
		{
			offset += SWAP68_W (de->rec_len);
			if (de->inode)
			{
				de1 = (ext2_d2 *) ((char *) de + EXT2_DIR_REC_LEN (de->name_len));
				de1->rec_len = SWAP68_W (SWAP68_W (de->rec_len) - EXT2_DIR_REC_LEN (de->name_len));
				de->rec_len = SWAP68_W (EXT2_DIR_REC_LEN (de->name_len));
				de = de1;
			}
			
			de->inode = 0;
			de->name_len = namelen;
			de->file_type = 0;
			memcpy (de->name, name, namelen);
			
			dir->in.i_mtime = dir->in.i_ctime = SWAP68_L (CURRENT_TIME);
			dir->in.i_flags = SWAP68_L (SWAP68_L (dir->in.i_flags) & ~EXT2_BTREE_FL);
			dir->in.i_version = SWAP68_L (++event);
			mark_inode_dirty (dir);
			bio_MARK_MODIFIED (&bio, u);
			
			clear_lastlookup (dir);
			
			*res_dir = de;
			*err = E_OK;
			
			return u;
		}
		
		offset += SWAP68_W (de->rec_len);
		de = (ext2_d2 *) ((char *) de + SWAP68_W (de->rec_len));
	}
	
	/* never reached */
	return NULL;
}

/*
 * ext2_delete_entry deletes a directory entry by merging it with the
 * previous entry
 */
long
ext2_delete_entry (ext2_d2 *dir, UNIT *u)
{
	ext2_d2 *de = (ext2_d2 *) u->data;
	ext2_d2 *pde = NULL;
	long i = 0;
	
	while (i < u->size)
	{
		if (!ext2_check_dir_entry ("ext2_delete_entry", NULL, de, u, i))
			return EIO;
		
		if (de == dir)
		{
			if (pde)
				pde->rec_len =
					SWAP68_W (SWAP68_W (pde->rec_len) +
						SWAP68_W (dir->rec_len));
			
			dir->inode = 0;
			
			bio_MARK_MODIFIED (&bio, u);
			return 0;
		}
		
		i += SWAP68_W (de->rec_len);
		pde = de;
		de = (ext2_d2 *) ((char *) de + SWAP68_W (de->rec_len));
	}
	
	return ENOENT;
}

# if 0
int ext2_mknod (struct inode * dir, struct dentry *dentry, int mode, int rdev)
{
	struct inode * inode;
	struct buffer_head * bh;
	struct ext2_dir_entry_2 * de;
	int err = -EIO;

	err = -ENAMETOOLONG;
	if (dentry->d_name.len > EXT2_NAME_LEN)
		goto out;

	inode = ext2_new_inode (dir, mode, &err);
	if (!inode)
		goto out;

	inode->i_uid = current->fsuid;
	inode->i_mode = mode;
	inode->i_op = NULL;
	bh = ext2_add_entry (dir, dentry->d_name.name, dentry->d_name.len, &de, &err);
	if (!bh)
		goto out_no_entry;
	de->inode = cpu_to_le32(inode->i_ino);
	dir->in.i_version = SWAP68_L (++event);
	if (S_ISREG(inode->i_mode)) {
		inode->i_op = &ext2_file_inode_operations;
		if (EXT2_HAS_INCOMPAT_FEATURE(dir->i_sb,
					      EXT2_FEATURE_INCOMPAT_FILETYPE))
			de->file_type = EXT2_FT_REG_FILE;
	} else if (S_ISDIR(inode->i_mode)) {
		inode->i_op = &ext2_dir_inode_operations;
		if (dir->i_mode & S_ISGID)
			inode->i_mode |= S_ISGID;
		if (EXT2_HAS_INCOMPAT_FEATURE(dir->i_sb,
					      EXT2_FEATURE_INCOMPAT_FILETYPE))
			de->file_type = EXT2_FT_DIR;
	}
	else if (S_ISLNK(inode->i_mode)) {
		inode->i_op = &ext2_symlink_inode_operations;
		if (EXT2_HAS_INCOMPAT_FEATURE(dir->i_sb,
					      EXT2_FEATURE_INCOMPAT_FILETYPE))
			de->file_type = EXT2_FT_SYMLINK;
	} else if (S_ISCHR(inode->i_mode)) {
		inode->i_op = &chrdev_inode_operations;
		if (EXT2_HAS_INCOMPAT_FEATURE(dir->i_sb,
					      EXT2_FEATURE_INCOMPAT_FILETYPE))
			de->file_type = EXT2_FT_CHRDEV;
	} else if (S_ISBLK(inode->i_mode)) {
		inode->i_op = &blkdev_inode_operations;
		if (EXT2_HAS_INCOMPAT_FEATURE(dir->i_sb,
					      EXT2_FEATURE_INCOMPAT_FILETYPE))
			de->file_type = EXT2_FT_BLKDEV;
	} else if (S_ISFIFO(inode->i_mode))  {
		init_fifo(inode);
		if (EXT2_HAS_INCOMPAT_FEATURE(dir->i_sb,
					      EXT2_FEATURE_INCOMPAT_FILETYPE))
			de->file_type = EXT2_FT_FIFO;
	}
	if (S_ISBLK(mode) || S_ISCHR(mode))
		inode->i_rdev = to_kdev_t(rdev);
	mark_inode_dirty(inode);
	mark_buffer_dirty(bh, 1);
	if (IS_SYNC(dir)) {
		ll_rw_block (WRITE, 1, &bh);
		wait_on_buffer (bh);
	}
	d_instantiate(dentry, inode);
	brelse(bh);
	err = 0;
out:
	return err;

out_no_entry:
	inode->i_nlink--;
	mark_inode_dirty(inode);
	iput(inode);
	goto out;
}
# endif
